/**
 * Represents a node in a doubly-linked list,
 * which can store data of type E.
 */
class ListNode<E> {
  private E data;
  public ListNode<E> next;
  public ListNode<E> prev;

  /**
   * Constructs a linked list node, saving the data.
   * The next and prev pointers are initially null.
   * 
   * @param data - data to store at this node.
   */
  public ListNode(E data) {
    this.data = data;
    prev = null;
    next = null;
  }

  /**
   * @return the data stored at this node.
   */
  public E get() {
    return data;
  }
}

/**
 * Represents a doubly-linked list,
 * where each node can store data of type E.
 */
public class LinkedList<E> {
  private ListNode<E> head;
  private ListNode<E> tail;
  private int size;

  /**
   * Creates an empty linked list.
   */
  public LinkedList() {
    head = null;
    tail = null;
    size = 0;
  }

  /**
   * Adds a node to the beginning of the list.
   * 
   * @param data - data to store at this node.
   */
  public void addFirst(E data) {
    ListNode<E> node = new ListNode<E>(data);
    node.next = head;
    if (head != null) {
      head.prev = node;
    } else {
      tail = node;
    }
    head = node;
    size++;
  }

  /**
   * Removes the node at the beginning of the list.
   */
  public void removeFirst() {
    if (head != null) {
      ListNode<E> next = head.next;
      if (next != null) {
        next.prev = null;
      } else {
        // next being null means we are removing the one and only item
        tail = null;
      }
      head = next;
      size--;
    }
  }

  /**
   * Retrieves the data stored at the node at a given index.
   * Assumes there is already a node at this index, i.e. index < size.
   * 
   * @param index - the index of the node to retrieve.
   * @return the data stored at the node at "index".
   */
  public E get(int index) {
    ListNode<E> node = head;
    for (int i = 0; i < index; i++) {
      node = node.next;
    }
    return node.get();
  }

  /**
   * @return Number of items in the list.
   */
  public int size() {
    return size;
  }

  /**
   * Adds a node at the end of the list.
   * 
   * @param data - data to store at the new last node.
   */
  public void addLast(E data) {
    ListNode<E> node = new ListNode<E>(data);
    if (tail == null) {
      // first item in list, also need to set the head node
      // could also check head == null or size == 0
      head = node;
    } else {
      tail.next = node;
      node.prev = tail;
    }
    tail = node;
    size++;
  }

  /**
   * Adds a node at the end of the list.
   * 
   * @param data - data to store at the new last node.
   */
  public void add(E data) {
    addLast(data);
  }

  /**
   * Removes the last node from the list.
   */
  public void removeLast() {
    if (tail != null) {
      ListNode<E> prev = tail.prev;
      if (prev == null) {
        head = null;
      } else {
        prev.next = null;
      }
      tail = prev;
      size--;
    }
  }

  /**
   * Adds a node at a given index, storing the requested data.
   * Any existing nodes from the node at the index to the end
   * will "shift" to the right.
   * Assumes there is a node at the index, i.e. index < size.
   * but also treats the cases at the ends, in which the list might be empty.
   * 
   * @param index - index of the node where to insert.
   * @param data  - data to store at the new node.
   */
  public void add(int index, E data) {

    if (index == 0) {
      // insert at the beginning
      addFirst(data);
      return;
    } else if (index == size) {
      // insert at the end
      addLast(data);
      return;
    }

    // insertion somewhere in between head and tail
    ListNode<E> node = new ListNode<E>(data);
    ListNode<E> prev = head;
    for (int i = 0; i < index - 1; i++) {
      prev = prev.next;
    }
    ListNode<E> next = prev.next;

    // next and prev should be non-null because of the checks at the
    // beginning of this method (index == 0 or size).
    prev.next = node;
    node.prev = prev;
    node.next = next;
    next.prev = node;

    size++;
  }

  /**
   * Removes a node at the index.
   * Assumes there is a node at the index, i.e. index < size.
   * 
   * @param index - index of the node to remove.
   */
  public void remove(int index) {
    if (head == null) {
      return; // empty list, nothing to remove
    } else if (index == 0) {
      removeFirst();
      return;
    } else if (index + 1 == size) {
      removeLast();
      return;
    }

    ListNode<E> node = head;
    for (int i = 0; i < index; i++) {
      node = node.next;
    }

    ListNode<E> prev = node.prev;
    ListNode<E> next = node.next;

    // prev and next are not null because we checked the index (0 or size - 1)
    // and called either removeFirst or removeLast accordingly
    prev.next = next;
    next.prev = prev;

    size--;
  }

  /**
   * @return a String representation of the linked list.
   */
  public String toString() {
    String result = "";
    ListNode<E> node = head;
    while (node != null) {
      result += node.get().toString();
      if (node.next != null)
        result += " <-> ";
      node = node.next;
    }
    return result;
  }

  /**
   * Tests the linked list implementation.
   * 
   * @param args
   */
  public static void main(String[] args) {
    LinkedList<Integer> list = new LinkedList<>();

    list.addFirst(5);
    list.addFirst(3);
    list.addFirst(2);
    // list.removeFirst();
    System.out.println(list);
    System.out.println(list.get(2));
    System.out.println("size = " + list.size());

    list.addLast(8);
    System.out.println(list);
    System.out.println("size = " + list.size());

    list.add(1, 4);
    System.out.println(list);
    System.out.println("size = " + list.size());

    list.remove(3);
    System.out.println(list);

    list.remove(3);
    System.out.println(list);

    list.remove(2);
    list.remove(1);
    list.remove(0);
    System.out.println(list);

    list.addFirst(5);
    list.addFirst(3);
    list.addFirst(2);
    System.out.println(list);
    for (int i = 0; i < 3; i++) {
      list.removeFirst();
      System.out.println(list);
    }

    list.addLast(9);
    System.out.println(list);

    list.remove(0);
    System.out.println(list);

    list.add(0, 4);
    list.add(0, 5);
    list.add(1, 6);
    list.add(3, 8);
    System.out.println(list);
  }
}